#!/usr/bin/env python
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------

"""Command line arguments."""

import argparse
import logging


def parse_arguments():
    """Parse CLI arguments."""
    parser = argparse.ArgumentParser(description='Anaconda Navigator.')
    parser.add_argument(
        '--version',
        help='Print version information',
        action='store_const',
        const=True,
        default=False,
        dest='version',
    )
    parser.add_argument(
        '--reset',
        help='Reset Navigator configuration',
        action='store_const',
        dest='reset',
        const=True,
        default=False,
    )
    parser.add_argument(
        '--remove-lock',
        help='Remove Navigator lock',
        action='store_const',
        dest='removelock',
        const=True,
        default=False,
    )
    parser.add_argument(
        '--verbose',
        help='Print information statements',
        action='store_const',
        dest='log_level',
        const=logging.INFO,
        default=logging.INFO,
    )
    parser.add_argument(
        '--debug',
        help='Print debug statements',
        action='store_const',
        dest='log_level',
        const=logging.DEBUG,
        default=logging.INFO,
    )

    return parser.parse_args()



# -*- coding: utf-8 -*-
"""
binaryornot.helpers
-------------------

Helper utilities used by BinaryOrNot.
"""

# yapf: disable

# Standard library imports
import logging

# Third party imports
import chardet


# yapf: enable

logger = logging.getLogger(__name__)


def print_as_hex(s):
    """
    Print a string as hex bytes.
    """
    print(":".join("{0:x}".format(ord(c)) for c in s))


def get_starting_chunk(filename, length=1024):
    """
    :param filename: File to open and get the first little chunk of.
    :param length: Number of bytes to read, default 1024.
    :returns: Starting chunk of bytes.
    """
    # Ensure we open the file in binary mode
    with open(filename, 'rb') as f:
        chunk = f.read(length)
        return chunk


_control_chars = b'\n\r\t\f\b'
_printable_ascii = _control_chars + bytes(range(32, 127))
_printable_high_ascii = bytes(range(127, 256))


def is_binary_string(bytes_to_check):
    """
    Uses a simplified version of the Perl detection algorithm,
    based roughly on Eli Bendersky's translation to Python:
    http://eli.thegreenplace.net/2011/10/19/perls-guess-if-file-is-text-or-binary-implemented-in-python/

    This is biased slightly more in favour of deeming files as text
    files than the Perl algorithm, since all ASCII compatible character
    sets are accepted as text, not just utf-8.

    :param bytes: A chunk of bytes to check.
    :returns: True if appears to be a binary, otherwise False.
    """

    # Empty files are considered text files
    if not bytes_to_check:
        return False

    # Now check for a high percentage of ASCII control characters
    # Binary if control chars are > 30% of the string
    low_chars = bytes_to_check.translate(None, _printable_ascii)
    nontext_ratio1 = float(len(low_chars)) / float(len(bytes_to_check))
    logger.debug('nontext_ratio1: %(nontext_ratio1)r', locals())

    # and check for a low percentage of high ASCII characters:
    # Binary if high ASCII chars are < 5% of the string
    # From: https://en.wikipedia.org/wiki/UTF-8
    # If the bytes are random, the chances of a byte with the high bit set
    # starting a valid UTF-8 character is only 6.64%. The chances of finding 7
    # of these without finding an invalid sequence is actually lower than the
    # chance of the first three bytes randomly being the UTF-8 BOM.

    high_chars = bytes_to_check.translate(None, _printable_high_ascii)
    nontext_ratio2 = float(len(high_chars)) / float(len(bytes_to_check))
    logger.debug('nontext_ratio2: %(nontext_ratio2)r', locals())

    is_likely_binary = (
        (nontext_ratio1 > 0.3 and nontext_ratio2 < 0.05) or (nontext_ratio1 > 0.8 and nontext_ratio2 > 0.8)
    )
    logger.debug('is_likely_binary: %(is_likely_binary)r', locals())

    # then check for binary for possible encoding detection with chardet
    detected_encoding = chardet.detect(bytes_to_check)
    logger.debug('detected_encoding: %(detected_encoding)r', locals())

    # finally use all the check to decide binary or text
    decodable_as_unicode = False
    if (detected_encoding['confidence'] > 0.9 and detected_encoding['encoding'] != 'ascii'):
        try:
            try:
                bytes_to_check.decode(encoding=detected_encoding['encoding'])
            except TypeError:
                # happens only on Python 2.6
                unicode(  # noqa
                    bytes_to_check, encoding=detected_encoding['encoding']
                )  # noqa
            decodable_as_unicode = True
            logger.debug('success: decodable_as_unicode: ' '%(decodable_as_unicode)r', locals())
        except LookupError:
            logger.debug('failure: could not look up encoding %(encoding)s', detected_encoding)
        except UnicodeDecodeError:
            logger.debug('failure: decodable_as_unicode: ' '%(decodable_as_unicode)r', locals())

    logger.debug('failure: decodable_as_unicode: ' '%(decodable_as_unicode)r', locals())
    if is_likely_binary:
        if decodable_as_unicode:
            return False
        else:
            return True
    else:
        if decodable_as_unicode:
            return False
        else:
            if b'\x00' in bytes_to_check or b'\xff' in bytes_to_check:
                # Check for NULL bytes last
                logger.debug('has nulls:' + repr(b'\x00' in bytes_to_check))
                return True
        return False



# -*- coding: utf8 -*-

"""Utilities for version comparison."""

from __future__ import annotations

__all__ = ['Version', 'compare', 'sort']

import collections.abc
import functools
import re
import typing

import packaging.version


NUMBER: typing.Final[re.Pattern[str]] = re.compile(r'([0-9]+)')


@functools.total_ordering
class Version:
    """Wrapper for versions strings that can be used for comparison."""

    def __init__(self, value: str | Version | None) -> None:
        """Initialize new instance of a :class:`~Version`."""
        self.raw: typing.Final[str] = str(value or '').strip()

    @functools.cached_property
    def parts(self) -> tuple[str | int, ...]:  # noqa: D401
        """
        Format-independent representation of a version.

        Might be used for a general comparison when version is not compatible with semver specifics.
        """
        return tuple(
            int(part) if (index % 2) else str(part)
            for index, part in enumerate(NUMBER.split(self.raw))
        )

    @functools.cached_property
    def version(self) -> packaging.version.Version | None:  # noqa: D401
        """Semver compatible version representation."""
        try:
            return packaging.version.Version(self.raw)
        except packaging.version.InvalidVersion:
            return None

    def __eq__(self, other: typing.Any) -> bool:
        """Check if instances are equal."""
        if isinstance(other, Version):
            return (self.raw == other.raw) or (self.version == other.version)
        return NotImplemented

    def __hash__(self) -> int:
        """Compute a hash of an instance."""
        return hash(self.raw)

    def __lt__(self, other: typing.Any) -> bool:
        """Check if current instance is less than other instance."""
        if isinstance(other, Version):
            if (self.version is not None) and (other.version is not None):
                return self.version < other.version
            return self.parts < other.parts
        return NotImplemented

    def __repr__(self) -> str:
        """Return string representation of an instance."""
        return f'{type(self).__name__}({self.raw!r})'

    def __str__(self) -> str:
        """Return string representation of an instance."""
        return self.raw


def as_version(value: str | Version) -> Version:
    """Make sure that :code:`value` is :class:`~Version`."""
    if isinstance(value, Version):
        return value
    return Version(value)


def compare(left: str | Version, right: str | Version) -> int:
    """Compare two versions."""
    left = as_version(left)
    right = as_version(right)
    if left == right:
        return 0
    if left < right:
        return -1
    return 1


@typing.overload
def sort(
        values: collections.abc.Iterable[str | Version],
        reverse: bool = ...,
        *,
        keep_versions: typing.Literal[False] = ...,
        unique: bool = ...,
) -> list[str]:
    """Sort sequence of versions."""


@typing.overload
def sort(
        values: collections.abc.Iterable[str | Version],
        reverse: bool = ...,
        *,
        keep_versions: typing.Literal[True],
        unique: bool = ...,
) -> list[Version]:
    """Sort sequence of versions."""


def sort(
        values: collections.abc.Iterable[str | Version],
        reverse: bool = False,
        *,
        keep_versions: bool = False,
        unique: bool = False,
) -> list[Version] | list[str]:
    """
    Sort sequence of versions.

    :param values: Versions to sort.
    :param reverse: Perform a reversed sorting (from largest to smallest).
    :param keep_versions: Return list of versions instead of list of strings.
    :param unique: Remove repeating versions from the output.
    :return: List of sorted versions.
    """
    if unique:
        values = set(values)

    result: list[Version] = sorted(map(as_version, values), reverse=reverse)

    if keep_versions:
        return result
    return list(map(str, result))
